#include <cstring>
#include <ctime>
#include <cctype>
#include <cstdio>
#include <cstdarg>
#include "utils.h"
#include "hashgen.h"

using namespace std;


#define commentHashTable \
  "Hash table. Each element either points to a key table entry,\n" \
  "or is null (%d) if there is no key with this hash value."

#define commentWeightTable \
  "Weight table (one character hash function)."

#define commentLengthTable \
  "Key lengths. Used for binary keys."

#define commentCollisionTable \
  "Collision table. Each element either points to a key with the same\n" \
  "hash value as the element's index, or is null (%d) if there are\n" \
  "no collisions."

#define commentKeyTable \
  "Key table. Holds keys and associated data (order preserved.)"

#define commentConstOffset \
  "Please notice that in expression `a[x-c]', where `a' is a static\n" \
  "array and `c' is a constant, `a-c' is a constant sub-expression,\n" \
  "so the offset should not incur additional runtime overhead."

#define commentCharRange \
  "Key characters must be in 0x%02x-0x%02x range."


void HashGenerator::generateLookupCode() {
  openOutput(outputHeaderName);
  doGenerateLookupCode();
  closeOutput();
  if (!outputTestName.empty()) {
    generateTestCode();
  }
}

void HashGenerator::doGenerateLookupCode() {
  adjustRanges();
  buildHashTable();
  generateDeclarations();
  generateDefinitions();
  delete[] hashTable;
}

void HashGenerator::adjustRanges() {
  if ((minExpectedChar == 0) || (minChar == 0)) {
    zeroTerm = false;
  }
  if (optimizeHits) {
    minHashValue = 0;
    hashValueRange = maxHashValue + 1;
  }
  needLengthTable = (!zeroTerm && (minKeyLen != maxKeyLen));
  firstChar = minExpectedChar + minOffset;
  if (optimizeHits) {
    int bytesBefore = hashValueRange * minSize(keyCount);
    if (collisionCount) {
      bytesBefore += keyCount * minSize(keyCount);
    }
    if (needLengthTable) {
      bytesBefore += keyCount * minSize(maxKeyLen);
    }
    firstChar += bytesBefore / minSize(maxWeight);
    if (firstChar > minExpectedChar) {
      firstChar = minExpectedChar;
    }
  }
  if (firstChar > minChar) {
    firstChar = minChar;
  }
  lastChar = maxExpectedChar + maxOffset;
  if (optimizeHits) {
    // Assumes that sizeof(char*) on target machine is at least 3.
    // If it's less, most likely there is no memory protection,
    // so it's okay anyway,
    int bytesAfter = keyCount * 3;
    lastChar -= bytesAfter / minSize(maxWeight);
    if (lastChar < maxExpectedChar) {
      lastChar = maxExpectedChar;
    }
  }
  if (lastChar < maxChar) {
    lastChar = maxChar;
  }
}

void HashGenerator::buildHashTable() {
  hashTable = new Key*[maxHashValue + 1];
  collisionCount = 0;
  for (int i = 0; i <= maxHashValue; i++) {
    hashTable[i] = 0;
  }
  for (Key* key = keys; key != end; key++) {
    key->collision = 0;
  }
  for (Key* key = keys; key != end; key++) {
    int hashValue = hash(key);
    Key* existing = hashTable[hashValue];
    hashTable[hashValue] = key;
    if (existing) {
      key->collision = existing;
      collisionCount++;
    }
  }
}

void HashGenerator::generateDeclarations() {
  generateLookupBanner();
  if (outputHeaderName != outputSourceName) {
    char buf[64];
    sprintf(buf, "%.50s_INCLUDED", lookupStructName.c_str());
    out("#ifndef %s\n", buf);
    out("#define %s\n", buf);
    out("\n");
  }
  if (includeString) {
    generateIncludes();
  }
  if ((language == langCPlusPlus) && !namespaceName.empty()) {
    out("namespace %s ", namespaceName.c_str());
    startScope();
    out("\n");
  }
  if (!declarations.empty()) {
    write(declarations.c_str());
  }
  else {
    out("struct %s ", entryStructName.c_str());
    startScope();
    out("char* %s;\n", keyName.c_str());
    endScope();
    out(";\n\n");
  }
  generateLookupStructDeclaration();
  if (!staticLookupData || (language == langCPlusPlus)) {
    generateLookupData(false);
  }
  if (inlineLookupFunc) {
    generateInline();
    generateLookupFunc(true);
  }
  else if (language != langCPlusPlus) {
    generateLookupFunc(false);
    out("\n");
  }
  if ((language == langCPlusPlus) && !namespaceName.empty()) {
    endScope(2);
  }
  if (outputHeaderName != outputSourceName) {
    out("#endif\n");
  }
}

void HashGenerator::generateDefinitions() {
  if (outputHeaderName != outputSourceName) {
    closeOutput();
    openOutput(outputSourceName);
    if (!outputHeaderName.empty()) {
      out("#include \"%s\"\n\n", outputHeaderName.c_str());
    }
  }
  if ((language == langCPlusPlus) && !namespaceName.empty()) {
    out("namespace %s ", namespaceName.c_str());
    startScope();
    out("\n");
  }
  if (!staticLookupData || (language == langCPlusPlus)) {
    generateLookupData();
  }
  if (!inlineLookupFunc) {
    generateLookupFunc();
  }
  if ((language == langCPlusPlus) && !namespaceName.empty()) {
    endScope(2);
  }
}

void HashGenerator::generateLookupBanner() {
  time_t sec = time(0);
  char buf[32];
  strcpy(buf, ctime(&sec));
  buf[strlen(buf) - 1] = 0;
  const char* languageName = "???";
  switch (language) {
    case langAnsiC    : languageName = "ANSI C"; break;
    case langOldC     : languageName = "K&R C" ; break;
    case langCPlusPlus: languageName = "C++"   ; break;
  };
  generateComment(
    "Generated by hashgen " VERSION " (" __DATE__ ")\n"
    "language: %s\n"
    "%d keys, hash table size %d (%d%%), %d collision(s)\n"
    "expected char range: 0x%02x-0x%02x\n"
    "%s",
    languageName,
    keyCount, hashValueRange,
    (hashValueRange * 100 + keyCount / 2) / keyCount,
    collisionCount,
    minExpectedChar, maxExpectedChar,
    buf
  );
  out("\n\n");
}

void HashGenerator::generateIncludes() {
  if (language == langCPlusPlus) {
    out("#include <cstring>\n\n");
  }
  else {
    out("#include <string.h>\n\n");
    if (language == langOldC) {
      out("#define const\n\n");
    }
  }
}

void HashGenerator::generateLookupStructDeclaration() {
  if (language == langCPlusPlus) {
    out("class %s ", lookupStructName.c_str());
    startScope();
    generateLabel("public");
    out("static size_t size() ");
    startScope();
    out("return %d;\n", keyCount);
    endScope(1);
    out("static const %s* at(size_t index) ", entryStructName.c_str());
    startScope();
    out("return %s + index;\n", keyTabName.c_str());
    endScope(1);
    if (zeroTerm) {
      out("static const %s* ", entryStructName.c_str());
      out("%s", lookupFuncName.c_str());
      out("(const char* key) ");
      startScope();
      out("return %s(key, std::strlen(key));\n", lookupFuncName.c_str());
      endScope(1);
    }
    out("static ");
    out(inlineLookupFunc ? "inline " : "");
    generateLookupFunc(false);
    generateLabel("protected");
  }
  else {
    out("struct %s ", lookupStructName.c_str());
    startScope();
  }
  generateHashTable(minHashValue, maxHashValue, false);
  if (collisionCount) {
    generateCollisionTable(false);
  }
  if (needLengthTable) {
    generateLengthTable(false);
  }
  generateWeightTable(firstChar, lastChar, false);
  generateKeyTable(false);
  endScope();
  out(";\n\n");
}

void HashGenerator::generateLookupData(bool aDef) {
  out(aDef ? "" : "extern ");
  if (language == langCPlusPlus) {
    out("%s %s;\n\n", lookupStructName.c_str(), lookupDataName.c_str());
    if (aDef) {
      generateHashTable(minHashValue, maxHashValue, true);
      if (collisionCount) {
        generateCollisionTable(true);
      }
      if (needLengthTable) {
        generateLengthTable(true);
      }
      generateWeightTable(firstChar, lastChar, true);
      generateKeyTable(true);
    }
  }
  else {
    out("struct %s %s", lookupStructName.c_str(), lookupDataName.c_str());
    if (aDef) {
      out(" = ");
      generateLookupDataInitializer();
    }
    out(";\n\n");
  }
}

void HashGenerator::generateLookupDataInitializer() {
  startScope();
  generateHashTableInitializer(minHashValue, maxHashValue);
  out(",\n");
  if (collisionCount) {
    generateCollisionTableInitializer(0, keyCount - 1);
    out(",\n");
  }
  if (needLengthTable) {
    generateLengthTableInitializer(0, keyCount - 1);
    out(",\n");
  }
  generateWeightTableInitializer(firstChar, lastChar);
  out(",\n");
  generateKeyTableInitializer(0, keyCount - 1);
  endScope();
}

void HashGenerator::generateTable(
  const char* aName,
  const char* aType,
  int aFrom, int aTo,
  void (HashGenerator::* aInit)(int, int))
{
  int size = aTo + 1 - aFrom;
  if (language == langCPlusPlus) {
    if (aInit) {
      out("%s %s::%s[%d]", aType, lookupStructName.c_str(), aName, size);
    }
    else {
      out("static %s %s[%d]", aType, aName, size);
    }
  }
  else {
    out("%s %s[%d]", aType, aName, size);
  }
  if (aInit) {
    out(" = ");
    (this->*aInit)(aFrom, aTo);
    out(";\n\n");
  }
  else {
    out(";\n");
  }
}

void HashGenerator::generateHashTable(int aFrom, int aTo, bool aDef) {
  generateTable(
    hashTabName.c_str(), minType(keyCount), aFrom, aTo,
    aDef ? &HashGenerator::generateHashTableInitializer : 0
  );
}

void HashGenerator::generateHashTableInitializer(int aFrom, int aTo) {
  startScope();
  int width = minWidth(keyCount);
  generateComment(commentHashTable, keyCount);
  for (int i = aFrom / 10 * 10; ; i++) {
    if (i % 10 == 0) {
      out("\n/* %*d */  ", minWidth(maxHashValue), i);
    }
    if (i >= aFrom) {
      out("%*u", width, hashTable[i] ? (hashTable[i] - keys) : keyCount);
      if (i == aTo) break;
      out(", ");
    }
    else {
      out("%*s", width + 2, "");
    }
  }
  endScope();
}

void HashGenerator::generateWeightTable(int aFrom, int aTo, bool aDef) {
  generateTable(
    weightTabName.c_str(), minType(maxWeight), aFrom, aTo,
    aDef ? &HashGenerator::generateWeightTableInitializer : 0
  );
}

void HashGenerator::generateWeightTableInitializer(int aFrom, int aTo) {
  startScope();
  int width = minWidth(maxWeight);
  generateComment(commentWeightTable);
  for (int i = aFrom / 10 * 10; ; i++) {
    if (i % 10 == 0) {
      out("\n/* %*d */  ", minWidth(aTo), i);
    }
    if (i >= aFrom) {
      out("%*u", width, weights[i]);
      if (i == aTo) break;
      out(", ");
    }
    else {
      out("%*s", width + 2, "");
    }
  }
  endScope();
}

void HashGenerator::generateLengthTable(bool aDef) {
  generateTable(
    lengthTabName.c_str(), minType(maxKeyLen), 0, keyCount - 1,
    aDef ? &HashGenerator::generateLengthTableInitializer : 0
  );
}

void HashGenerator::generateLengthTableInitializer(int, int) {
  startScope();
  int width = minWidth(maxKeyLen);
  generateComment(commentLengthTable);
  for (int i = 0; ; ) {
    if (i % 10 == 0) {
      out("\n/* %*d */  ", minWidth(keyCount), i);
    }
    out("%*u", width, keys[i].keyLen);
    if (++i >= keyCount) break;
    out(", ");
  }
  endScope();
}

void HashGenerator::generateCollisionTable(bool aDef) {
  generateTable(
    collisionTabName.c_str(), minType(keyCount), 0, keyCount - 1,
    aDef ? &HashGenerator::generateCollisionTableInitializer : 0
  );
}

void HashGenerator::generateCollisionTableInitializer(int, int) {
  startScope();
  int width = minWidth(keyCount);
  generateComment(commentCollisionTable, keyCount);
  for (int i = 0; ; ) {
    if (i % 10 == 0) {
      out("\n/* %*d */  ", width, i);
    }
    out("%*u",
      width,
      keys[i].collision ? (keys[i].collision - keys) : keyCount
    );
    if (++i >= keyCount) break;
    out(", ");
  }
  endScope();
}

void HashGenerator::generateKeyTable(bool aDef) {
  string name = (language == langCPlusPlus) ? "" : "struct ";
  name += entryStructName;
  generateTable(
    keyTabName.c_str(), name.c_str(), 0, keyCount - 1,
    aDef ? &HashGenerator::generateKeyTableInitializer : 0
  );
}

void HashGenerator::generateKeyTableInitializer(int, int) {
  startScope();
  generateComment(commentKeyTable);
  int maxLen = maxKeyLen + 4;
  for (int i = 0; ; ) {
    out("\n/* %*u */  { ", minWidth(keyCount), i);
    char buf[1024];
    int len = strQuote((char*)(keys[i].key), keys[i].keyLen, buf);
    if (len > maxLen) {
      maxLen = len;
    }
    out("%s", buf);
    if (*keys[i].data) {
      out(",%*s %s", maxLen - len, "", keys[i].data);
    }
    if (++i >= keyCount) {
      out(" }");
      break;
    }
    out(" },");
  }
  endScope();
}

void HashGenerator::generateInline() {
  switch (language) {
    case langAnsiC:
      out(
        "#if defined(__cplusplus)\n"
        "  inline\n"
        "#elif defined(__GNUC__)\n"
        "  __inline__\n"
        "#elif defined(INLINE)\n"
        "  INLINE\n"
        "#endif\n\n"
      );
      break;
    case langOldC:
      out(
        "#ifdef __GNUC__\n"
        "  __inline__\n"
        "#else\n"
        "#ifdef INLINE\n"
        "  INLINE\n"
        "#endif\n"
        "#endif\n\n"
      );
      break;
    case langCPlusPlus:
      break;
  }
}

void HashGenerator::generateHashUpdate(int aPos, int aHashPos) {
  out("h += %s.%s", lookupDataName.c_str(), weightTabName.c_str());
  out("[(unsigned char)key[%d]", aPos);
  out(firstChar ? "%+d" : "", -firstChar);
  out(offsets[aHashPos] ? "%+d" : "", offsets[aHashPos]);
  constOffsetNote |= (firstChar || offsets[aHashPos]);
  out("];\n");
}

void HashGenerator::generateLookupFunc(bool aDef) {
  if (aDef && ((minExpectedChar > 1) || (maxExpectedChar < 255))) {
    generateComment(commentCharRange, minExpectedChar, maxExpectedChar);
    out("\n\n");
  }
  if (language == langCPlusPlus) {
    out("%s* ", entryStructName.c_str());
    if (aDef) {
      out("%s::", lookupStructName.c_str());
    }
    out("%s", lookupFuncName.c_str());
    out("(const char* key, size_t len)");
  }
  else {
    out("struct %s* %s", entryStructName.c_str(), lookupFuncName.c_str());
    if (language == langOldC) {
      if (aDef) {
        out("(key, len)\n  register char* key;\n  register unsigned len;\n");
      }
      else {
        out("(/* char* key, unsigned len */)");
      }
    }
    else {
      out("(const char* key, size_t len)");
    }
  }
  if (aDef) {
    out((language == langOldC) ? "" : " ");
    generateLookupFuncDefinition();
  }
  else {
    out(";\n");
  }
}

void HashGenerator::generateCompare() {
  out("if (");
  if (zeroTerm) {
    out("*key == *p->%s && ", keyName.c_str());
    out("!strcmp(key, p->%s)", keyName.c_str());
  }
  else {
    if (minKeyLen != maxKeyLen) {
      out("len == %s.%s[h] && ", lookupDataName.c_str(), lengthTabName.c_str());
    }
    out("!memcmp(key, p->name, len)");
  }
  out(")\n  return p;\n");
}

void HashGenerator::generateLookupFuncDefinition() {
  startScope();
  constOffsetNote = false;
  const char* null = "NULL";
  if (language == langCPlusPlus) {
    null = "0";
    out("using namespace std;\n");
  }
  else if (staticLookupData) {
    out("static ");
    generateLookupData(true);
  }
  out("register struct %s* p;\n", entryStructName.c_str());
  out("register unsigned h = (unsigned)len;\n");
  if (lengthSwitch && (minKeyLen != maxKeyLen)) {
    if (!optimizeHits) {
      out("if (h < %d)\n  return %s;\n", minKeyLen, null);
    }
    out("switch (h) ");
    startScope();
    out("default: return %s;\n", null);
    int pos = maxHashKeyLen - 1;
    for (int len = maxKeyLen; len >= 0; len--) {
      out("case %*d: ", minWidth(hashPosMap[maxHashKeyLen - 1]), len);
      if ((pos >= 0) && (hashPosMap[pos] == len - 1)) {
        out("\n  ");
        generateHashUpdate(len - 1, pos);
        pos--;
      }
      else {
        out((len <= hashPosMap[0]) ? ";\n" : "\n");
      }
    }
    endScope(1);
  }
  else {
    if (minKeyLen == maxKeyLen) {
      out("if (h != %d)\n  return %s;\n", maxKeyLen, null);
    }
    else if (minKeyLen) {
      out("if (h < %d", minKeyLen);
      out(!optimizeHits ? " || h > %d" : "", maxKeyLen);
      out(")\n  return %s;\n", null);
    }
    else if (!optimizeHits) {
      out("if (h > %d)\n  return %s;\n", maxKeyLen, null);
    }
    bool checkLabel = false;
    for (int i = 0; i < maxHashKeyLen; i++) {
      int pos = hashPosMap[i];
      if (pos > minKeyLen) {
        out("if (len < %d)\n  goto check;\n", pos + 1);
        checkLabel = true;
      }
      generateHashUpdate(pos, i);
    }
    if (checkLabel) {
      generateLabel("check");
    }
  }
  out("if (");
  out(minHashValue ? "h < %d || " : "", minHashValue);
  out("h > %d)\n  return %s;\n", maxHashValue, null);
  out("h = %s.%s[h", lookupDataName.c_str(), hashTabName.c_str());
  out(minHashValue ? "%+d" : "", -minHashValue);
  out("];\n");
  constOffsetNote |= (minHashValue != 0);
  if (collisionCount) {
    out("while (h != %d) ", keyCount);
    startScope();
  }
  else if (hashValueRange > keyCount) {
    out("if (h == %d)\n  return %s;\n", keyCount, null);
  }
  out("p = &%s.%s[h];\n", lookupDataName.c_str(), keyTabName.c_str());
  generateCompare();
  if (collisionCount) {
    out("h = %s.%s[h];\n", lookupDataName.c_str(), collisionTabName.c_str());
    endScope(1);
  }
  out("return %s;", null);
  endScope(2);
  if (constOffsetNote) {
    generateComment(commentConstOffset);
    out("\n\n");
  }
}

void HashGenerator::generateLabel(const char* aName) {
  unindent();
  out("%s:\n", aName);
  indent();
}

void HashGenerator::generateComment(const char* aFormat, ...) {
  char buf[writeBufSize];
  va_list arg;
  va_start(arg, aFormat);
  vsprintf(buf, aFormat, arg);
  va_end(arg);
  if (language == langCPlusPlus) {
    write("// ");
    linePrefix = "// ";
  }
  else {
    write("/* ");
    linePrefix = " * ";
  }
  write(buf);
  linePrefix = "";
  if (language != langCPlusPlus) {
    write(" */");
  }
}
void HashGenerator::startScope() {
  out("{\n");
  indent();
}

void HashGenerator::endScope(int aLines) {
  unindent();
  if (outputColumn) {
    out("\n");
  }
  out("}");
  while (aLines--) {
    out("\n");
  }
}


