#include <cstring>
#include <climits>
#include <cstdlib>
#include <ctime>
#include <algorithm>
#include <functional>

#include "utils.h"
#include "hashgen.h"

using namespace std;


struct inSet {
  bool* set;
  inSet(bool* aSet) : set(aSet) {}
  bool operator()(const Key& aKey) {
    const short* key = aKey.hashKey;
    bool result = true;
    for (int len = aKey.hashKeyLen; len && result; len--) {
      result = set[*key++];
    }
    return result;
  }
};

static void addToSet(const Key& aKey, bool* aSet) {
  const short* key = aKey.hashKey;
  for (int len = aKey.hashKeyLen; len; len--) {
    aSet[*key++] = true;
  }
}

inline bool moreFreq(const Key& aKey1, const Key& aKey2) {
  return aKey1.freq > aKey2.freq;
}

void HashGenerator::reorder() {
  bool covered[maxCharCount];
  fill(covered, covered + maxCharCount, false);
  for (Key* key = keys; key != end; ) {
    Key* max = min_element(key, end, moreFreq);
    swap(*max, *key);
    addToSet(*key++, covered);
    Key* rest = partition(key, end, inSet(covered));
    sort(key, rest, moreFreq);
    key = rest;
  }
}

bool HashGenerator::changeWeight(Key* aKey, int aChar) {
  // Try to change aChar's weight to resolve collision
  // between aKey and its predecessors.
  int saveWeight = weights[aChar];
  int index = aChar >> 3;
  byte mask = 1 << (aChar & 7);
  for (int count = maxWeight; count; count--) {
    if (++weights[aChar] >= maxWeight) {
      weights[aChar] = 0;
    }
    // Rehash with the new weight trying to get less collisions.
    HashSet::reset();
    int collisions = 0;
    for (Key* key = keys; ; ) {
      // Avoid rehashing for keys that don't depend on aChar.
      int h = (key->charSet[index] & mask) ? hash(key) : key->hash;
      if (HashSet::check(h)) { // check(): sets & returns previous state
        if (++collisions >= collisionCount) {
          break;
        }
      }
      if (key++ == aKey) {
        collisionCount = collisions;
        return true;
      }
    }
  }
  weights[aChar] = saveWeight;
  return false;
}

struct indirectCompare {
  int* table;
  explicit indirectCompare(int* aTable) : table(aTable) {}
  bool operator()(int a, int b) const { return table[a] < table[b]; }
};

bool HashGenerator::resolveCollision(Key* aPrior, Key* aKey) {
  short dif[maxHashKey * 2];
  short* difEnd = set_symmetric_difference(
    aKey->sortedHashKey, aKey->sortedHashKey + aKey->hashKeyLen,
    aPrior->sortedHashKey, aPrior->sortedHashKey + aPrior->hashKeyLen,
    dif
  );
  sort(dif, difEnd, indirectCompare(charFreq));
  for (short* difChar = dif; difChar != difEnd; difChar++) {
    if (changeWeight(aKey, *difChar)) {
      return true;
    }
  }
  if (debug) {
    char buf[maxLine];
    strQuote(reinterpret_cast<char*>(aKey->key), aKey->keyLen, buf);
    log("unresolved collision at %d, %s/", aKey - keys, buf);
    strQuote(reinterpret_cast<char*>(aPrior->key), aPrior->keyLen, buf);
    log("%s/%d\n", buf, difEnd - dif);
  }
  for (Key* key = keys; key != aKey + 1; key++) {
    hash(key);
  }
  return false;
}

void HashGenerator::computeWeights() {
  if (debug) {
    log("max weight %d%s\n", maxWeight, random ? ", random" : "");
  }
  if (random) {
    generate(
      weights,
      weights + maxCharCount,
      Random(randomGenerator, maxWeight)
    );
  }
  else {
    fill(
      weights,
      weights + maxCharCount,
      0
    );
  }
  HashSet::setSize(maxHashTabSize);
  collisionCount = 0;
  for (Key* key = keys; key != end; key++) {
    hash(key);
    for (Key* prior = keys; prior != key; prior++) {
      if (prior->hash == key->hash) {
        collisionCount++;
        bool resolved = resolveCollision(prior, key);
        if (!resolved && (duplicateCount == 0)) {
	  while (!resolved && (maxWeight < maxWeightLimit)) {
	    maxWeight = min(maxWeightLimit, maxWeight + 1 + maxWeight / 2);
	    if (debug) {
	      log("increased max weight to %d\n", maxWeight);
	    }
	    resolved = resolveCollision(prior, key);
	  }
        }
        if (!resolved && debug) {
          log("failed to resolve collision\n");
        }
        break;
      }
    }
  }
  HashSet::setSize(0);
}

void HashGenerator::computeHashRange() {
  minHashValue = INT_MAX;
  maxHashValue = 0;
  for (Key* key = keys; key != end; key++) {
    int h = hash(key); // just in case, key->hash should be valid
    maxHashValue = max(maxHashValue, h);
    minHashValue = min(minHashValue, h);
  }
  hashValueRange = maxHashValue + 1 - minHashValue;
}

// Ensures that a key containing in some hash position a character
// that doesn't occur in the same position in any of the valid keywords
// will get a very big hash value (in order to reduce average
// unsuccessful search time).

void HashGenerator::boostUnusedWeights() {
  maxWeight = *max_element(weights, weights + maxCharCount);
  if (maxWeight <= 255) {
    maxWeight = 255;
  }
  else if (maxWeight <= 9999) { // with 65535 the weight table's code
    maxWeight = 9999;           // doesn't fit in 80 columns  :)
  }
  for (int charCode = 0; charCode < maxCharCount; charCode++) {
    if (charFreq[charCode] == 0) {
      weights[charCode] = maxWeight;
    }
  }
}

bool originalOrdering(const Key& aKey1, const Key& aKey2) {
  return aKey1.no < aKey2.no;
}

void HashGenerator::run() {
  randomGenerator.set(randomize ? static_cast<long>(time(0)) : 49157);
  makeHashKeys();
  checkExpectedCharRange();
  computeCharOffsets();
  gatherStatistics();
  reorder();
  maxWeight = 2 + keyCount / 2 + sizeFactor * (keyCount / 8);
  maxWeightLimit = maxWeight * 8;
  maxHashTabSize = maxKeyLen + maxHashKeyLen * maxWeightLimit;
  computeWeights();
  boostUnusedWeights();
  computeHashRange();
  sort(keys, end, originalOrdering);
  log("hash value range %d ", hashValueRange);
  log("(%d%%), ", hashValueRange * 100 / keyCount);
  log("collisions %d\n", collisionCount);
  generateLookupCode();
}


