/*

Tests some common hash functions.
Has nothing to do with hashgen.

Synopsis: hashbench [filename]

If no filename is specified, the standard input is read.
The input contains a list of keywords, each on a new line.

Examples.

Webster dictionary (200,000+ words):

    ./hashbench /usr/share/dict/words

Symbols of a program `foo':

    nm /usr/local/bin/foo | cut -c 12-80 | sort -u | ./hashbench


Tested on English words, proper names, symbol tables of large 
executables. The bottom line:

All tested hash functions dispersed the keys almost equally well.

ROT5 is probably the best.

ELF (PJW) is good, but not so good as they say it is; simpler
(faster) functions yield even better results.

CRC32 is one of the slowest, but may be good for external hashing.

Tweaked versions can be 10-20% faster than the readable ones.
With a good optimizing compiler recoding in assembly isn't worth
the effort. 

*/

#include <ctime>
#include <climits>
#include <cstdio>
#include <cassert>
#include <string>
#include <vector>
#include <algorithm>
#include <functional>
#include <fstream>

using namespace std;


#if UINT_MAX == 0xFFFF
  typedef unsigned long uint32;
#else
  typedef unsigned int uint32;
#endif

typedef unsigned char byte;


#define HASH_FUNC(NAME, UPDATE) \
                       \
  uint32 NAME(const char* key, size_t) { \
    uint32 h = 0;      \
    while (*key) {     \
      UPDATE;          \
    }                  \
    return h;          \
  }                    \
                       \
  uint32 NAME##Fast(const char* key, size_t len) { \
    /* Classical Duff's Device would have been shorter, but it's too ugly */ \
    uint32 h = 0;      \
    while (len >= 8) { \
      UPDATE; UPDATE;  \
      UPDATE; UPDATE;  \
      UPDATE; UPDATE;  \
      UPDATE; UPDATE;  \
      len -= 8;        \
    }                  \
    /* Jump table */   \
    switch (len) {     \
      case 7: UPDATE;  \
      case 6: UPDATE;  \
      case 5: UPDATE;  \
      case 4: UPDATE;  \
      case 3: UPDATE;  \
      case 2: UPDATE;  \
      case 1: UPDATE;  \
    }                  \
    return h;          \
  } 


// Hash functions.


HASH_FUNC(hashRot3,  h = ((h << 3) | (h >> 29)) + byte(*key++))
HASH_FUNC(hashRot5,  h = ((h << 5) | (h >> 27)) + byte(*key++))
HASH_FUNC(hashMul19, h = h * 19 + byte(*key++))
HASH_FUNC(hashMul33, h = h + (h << 5) + byte(*key++))
HASH_FUNC(hashSdbm,  h = (h << 6) + (h << 16) - h + byte(*key++))

HASH_FUNC(hashElf, {
    h = (h << 4) + byte(*key++);
    uint32 highBits = h & 0xF0000000;
    if (highBits) {
      h = (h ^ (highBits >> 24)) & 0xFFFFFFF;
    }
  })

uint32 crc32Table[256];

HASH_FUNC(hashCrc32, h = (h << 8) ^ crc32Table[byte((h >> 24) ^ byte(*key++))])

void initCrc32() {
  uint32 i, j, c;
  for (i = 0; i < 256; i++) {
    for (c = i << 24, j = 8; j > 0; --j) {
      c = c & 0x80000000 ? (c << 1) ^ 0x04c11db7 : (c << 1);
    }
    crc32Table[i] = c;
  }
}


// A reference point for timing. Includes 
// * The function call overhead.
// * References the arguments mostly to ensure that the usual stack frame
//   code is not optimized out of existence.
// * Returns increasing values to facilitate CPU cache misses while testing 
//   since it's common for all hash functions.

uint32 hashDummy(const char* key, size_t len) {
  static uint32 c = 0;
  return ++c + *key + len;
}


class Timer {
public:
  Timer() { ref = clock(); }
  clock_t ticksElapsed() const { return clock() - ref; }
  double secsElapsed() const { return 1.0 * (clock() - ref) / ticksPerSec(); }
  static clock_t ticksPerSec() { return CLK_TCK; }
  // Avoids using sysconf() to stay portable.
private:
  clock_t ref;
};

uint32 nextPrime(uint32 n) {
  static const uint32 primes[] = {
    53ul,         97ul,         193ul,       389ul,       769ul,
    1543ul,       3079ul,       6151ul,      12289ul,     24593ul,
    49157ul,      98317ul,      196613ul,    393241ul,    786433ul,
    1572869ul,    3145739ul,    6291469ul,   12582917ul,  25165843ul,
    50331653ul,   100663319ul,  201326611ul, 402653189ul, 805306457ul, 
    1610612741ul, 3221225473ul, 4294967291ul
  };
  const uint32* last = primes + sizeof(primes) / sizeof(primes[0]);
  const uint32* pos = lower_bound(primes, last, n);
  return (pos == last) ? *(last - 1) : *pos;
}


vector<string> keys;
vector<long> counts;
long modulo;
long dummyTime;


void loadKeys(istream& stream) {
  keys.clear();
  char buf[256];
  while (stream.getline(buf, sizeof(buf))) {
    keys.push_back(buf); 
  }
  if (keys.empty()) {
    cerr << "Nothing to do\n";
    exit(1);
  }
  random_shuffle(keys.begin(), keys.end());
  modulo = nextPrime(keys.size() * 3 / 2);
  counts.resize(modulo);
  cout << "  " << keys.size() << " keys, hash table size " << modulo << "\n";
  cout << "  hashfunc   fill cavg cmax   ns\n";
  cout << "  ---------- ---- ---- ---- ----\n";
}

void loadKeys(const char* fileName) {
  if (fileName) {
    ifstream stream(fileName);
    if (!stream) {
      cerr << "Cannot open " << fileName << "\n";
      exit(1);
    }
    cout << fileName << "\n";
    loadKeys(stream);
  }
  else {
    loadKeys(cin);
  }
}

long testRun(uint32 (*hash)(const char*, size_t)) {
  fill_n(counts.begin(), modulo, 0);
  unsigned n = 0;
  Timer timer;
  do {
    vector<string>::const_iterator i = keys.begin(); 
    for ( ; i != keys.end(); ++i) {
      unsigned pos = hash(i->c_str(), i->length()) % modulo;
      ++counts[pos];
    }
    ++n;
  } while ((n < UINT_MAX) && (timer.ticksElapsed() < timer.ticksPerSec()));
  long time = long(timer.secsElapsed() * 1e9 / (1.0 * n * keys.size()));
  transform(
    counts.begin(), counts.end(), 
    counts.begin(), 
    bind2nd(divides<uint32>(), n)
  ); 
  return time;
}

long testRunAvg(uint32 (*hash)(const char*, size_t)) {
  const int n = 8; // even  
  long times[n];
  for (int i = 0; i < n; i++) {
    times[i] = testRun(hash);
  }
  // Throws away the best and the worst results
  // and averages 2 median values.
  sort(times, times + n);
  return (times[n / 2 - 1] + times[n / 2]) / 2;
}

void test(uint32 (*hash)(const char*, size_t), const char* name) {
  long hashTime = max(testRunAvg(hash) - dummyTime, 0L);
  double avgChain = 0;
  long slotsUsed = 0;
  long longestChain = 0;
  vector<long>::const_iterator i = counts.begin(); 
  for ( ; i != counts.end(); ++i) {
    long count = *i;
    if (count) {
      slotsUsed++;
      avgChain += count;
      if (count > longestChain) {
        longestChain = count;
      }
    }
  }
  avgChain /= slotsUsed;
  char buf[256];
  sprintf(
    buf, 
    "  %-10s %3ld%% %4.1f %4ld %4ld\n",
    name, slotsUsed * 100 / modulo, 
    avgChain, longestChain, hashTime
  );
  cout << buf;
}

void test(const char* fileName = 0) {
  assert(sizeof(uint32) == 4);
  loadKeys(fileName);
  dummyTime = testRunAvg(hashDummy);
  initCrc32();
  test(hashRot3,      "rot3");
  test(hashRot5,      "rot5");
  test(hashMul19,     "mul19");
  test(hashMul33,     "mul33");
  test(hashSdbm,      "sdbm");
  test(hashElf,       "elf");
  test(hashCrc32,     "crc32");
  test(hashRot3Fast,  "rot3 opt");
  test(hashRot5Fast,  "rot5 opt");
  test(hashMul19Fast, "mul19 opt");
  test(hashMul33Fast, "mul33 opt");
  test(hashSdbmFast,  "sdbm opt");
  test(hashElfFast,   "elf opt");
  test(hashCrc32Fast, "crc32 opt");
}

int main(int argc, char* argv[]) {
  if (argc <= 1) {
    test(); // stdin
  }
  else {
    for (int i = 1; i < argc; i++) { // file(s)
      test(argv[i]);
    }
  }
  return 0;
}

