#ifndef _KEYSET_H_
#define _KEYSET_H_

#include <fstream>
#include <string>
#include <stdexcept>

#include "types.h" 
#include "rand.h" 
#include "fastset.h" 
#include "keylist.h" 


class HashSet : public FastSet<short> { };


class HashGenerator : private KeyList, private HashSet {
  // Inherits from HashSet only for efficiency
  // (normally should have contained it).
public:

  struct Exception : public std::logic_error { 
    Exception(const char* aMessage) : std::logic_error(aMessage) {}
  };

  enum { // enum instead of static const for portability reasons
    maxHashKey = 32,
    maxLine = 8 * 1024,
    maxCharCount = maxHashKey * 256,
  };
  
  HashGenerator(const char* aArgs[]);
  ~HashGenerator();

  void run();

private:

  RandomGenerator randomGenerator;

  // Algorithm

  int duplicateCount;
  int minHashKeyLen;
  int maxHashKeyLen;
  int minChar;
  int maxChar;
  int charFreq[maxCharCount];
  int hashPosMap[maxHashKey];
  int offsets[maxHashKey];
  int minOffset;
  int maxOffset;

  int maxWeight;
  int maxWeightLimit;
  int maxHashTabSize;

  int collisionCount;
  int weights[maxCharCount];

  int minHashValue;
  int maxHashValue;
  int hashValueRange;
  Key** hashTable;

  // Input

  const char** args;
  std::string inputFileName;
  std::istream* inputStream;
  std::ifstream inputFile;

  // Output

  enum { writeBufSize = 8 * 1024 };
  std::ostream* outputStream;
  std::ofstream outputFile;
  int outputIndent;
  int outputColumn;
  std::string linePrefix;

  // Code generation

  int firstChar;
  int lastChar;
  bool needLengthTable;
  bool constOffsetNote;

  // Params 

  bool random;
  bool randomize;
  int sizeFactor;

  bool zeroTerm;
  bool optimizeHits;
  bool lengthSwitch;
  int minExpectedChar;
  int maxExpectedChar;

  enum Language {
    langAnsiC, 
    langOldC, 
    langCPlusPlus 
  };

  Language language;
  bool inlineLookupFunc;
  bool staticLookupData;
  bool includeString;

  std::string outputSourceName;
  std::string outputHeaderName;
  std::string outputTestName;
  
  std::string namespaceName;
  std::string entryStructName;
  std::string keyName;
  std::string lookupStructName; 
  std::string lookupDataName; 
  std::string hashTabName; 
  std::string weightTabName; 
  std::string keyTabName; 
  std::string lengthTabName; 
  std::string collisionTabName; 
  std::string lookupFuncName; 

  std::string declarations;

  bool debug;

  uint keyDifMask(const byte*, int, const byte*, int);
  void parseInput();
  void parseEntry(int aLine, const char* aBuf);
  void parseOption(int aLine, const char* aBuf);
  void showUsage();
  void setOption(int aLine, const char* aName, const char* aValue);
  void setDefaultOptions();
  void parseArgs();
  void openInput();
  void closeInput();
  uint findRedundantPositions();
  uint findRedundantPositionsFast();
  uint findRedundantPositionsSlow();
  void makeHashKeys();
  void computeCharOffsets();
  bool findSimilarHashKeys();
  void applyOffsets();
  void checkExpectedCharRange();
  void gatherStatistics();
  void reorder();
  void computeHashRange();
  void computeWeights();
  void boostUnusedWeights();
  bool resolveCollision(Key*, Key*);
  bool changeWeight(Key*, int aChar);
  void buildHashTable();
  void adjustRanges();
  inline int hash(Key*);

  void out(const char*, ...);
  void log(const char*, ...);
  void die(const char*, ...);
  void comment(const char*, ...);
  void write(const char*);
  void openOutput(const std::string& aFileName);
  void closeOutput();
  void indent() { outputIndent++; }
  void unindent() { outputIndent--; }

  typedef void (HashGenerator::*TableInit)(int, int);

  void startScope();
  void endScope(int aNewLines = 0);
  void generateLookupCode();
  void doGenerateLookupCode();
  void generateLookupBanner();
  void generateDeclarations();
  void generateDefinitions();
  void generateLookupStructDeclaration();
  void generateLookupData(bool aDef = true);
  void generateLookupDataInitializer();
  void generateTable(const char*, const char*, int, int, TableInit);
  void generateHashTable(int aFrom, int aTo, bool aDef = true);
  void generateHashTableInitializer(int aFrom, int aTo);
  void generateWeightTable(int aFrom, int aTo, bool aDef = true);
  void generateWeightTableInitializer(int aFrom, int aTo);
  void generateKeyTable(bool aDef = true);
  void generateKeyTableInitializer(int aFrom, int aTo);
  void generateLengthTable(bool aDef = true);
  void generateLengthTableInitializer(int aFrom, int aTo);
  void generateCollisionTable(bool aDef = true);
  void generateCollisionTableInitializer(int aFrom, int aTo);
  void generateLookupFunc(bool aDef = true);
  void generateLookupFuncDefinition();
  void generateInline();
  void generateIncludes();
  void generateCompare();
  void generateHashUpdate(int aPos, int aHashPos);
  void generateLabel(const char*);
  void generateComment(const char*, ...);
  void generateCompare(const char*);

  void generateTestBanner();
  void generateTestIncludes();
  void generateTestKey(const byte*, int, bool);
  void generateTestKeys(bool);
  void generateTestErrorBranch(const char*, const char*);
  void generateTestMain();
  void generateTestCode();
};

// hash(key): Calculates hash value of a key.
// Time critical (though way too long too be inlined).

int HashGenerator::hash(Key* aKey) {
  register short* key = aKey->hashKey;
  register int h = aKey->keyLen;
  // Most compilers will generate a jump table
  switch (aKey->hashKeyLen) {  
    default:
    case 32: h += weights[key[31]];
    case 31: h += weights[key[30]];
    case 30: h += weights[key[29]];
    case 29: h += weights[key[28]];
    case 28: h += weights[key[27]];
    case 27: h += weights[key[26]];
    case 26: h += weights[key[25]];
    case 25: h += weights[key[24]];
    case 24: h += weights[key[23]];
    case 23: h += weights[key[22]];
    case 22: h += weights[key[21]];
    case 21: h += weights[key[20]];
    case 20: h += weights[key[19]];
    case 19: h += weights[key[18]];
    case 18: h += weights[key[17]];
    case 17: h += weights[key[16]];
    case 16: h += weights[key[15]];
    case 15: h += weights[key[14]];
    case 14: h += weights[key[13]];
    case 13: h += weights[key[12]];
    case 12: h += weights[key[11]];
    case 11: h += weights[key[10]];
    case 10: h += weights[key[ 9]];
    case  9: h += weights[key[ 8]];
    case  8: h += weights[key[ 7]];
    case  7: h += weights[key[ 6]];
    case  6: h += weights[key[ 5]];
    case  5: h += weights[key[ 4]];
    case  4: h += weights[key[ 3]];
    case  3: h += weights[key[ 2]];
    case  2: h += weights[key[ 1]];
    case  1: h += weights[key[ 0]];
    case  0: ;
  }
  aKey->hash = h;
  return h;
}

#endif

