#include <cstring>
#include <cctype>
#include <climits>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <algorithm>

#include "utils.h"
#include "options.h"
#include "hashgen.h"

using namespace std;


HashGenerator::HashGenerator(const char* aArgs[]) {
  args = aArgs;
  setDefaultOptions();
  parseInput();
}

HashGenerator::~HashGenerator() {
}

void HashGenerator::setDefaultOptions() {
  optimizeHits = false;
  random = false;
  randomize = false;
  sizeFactor = 0;
  minExpectedChar = 1;
  maxExpectedChar = 255;
  entryStructName   = "entry";
  keyName           = "name";
  namespaceName     = "";
  lookupStructName  = "lookup_struct";
  lookupDataName    = "lookup_data";
  hashTabName       = "hash";
  keyTabName        = "key";
  weightTabName     = "weight";
  lengthTabName     = "length";
  collisionTabName  = "collision";
  lookupFuncName    = "lookup";
  outputHeaderName  = "";
  outputSourceName  = "";
  outputTestName    = "";
  inputFileName     = "";
  language          = langAnsiC;
  inlineLookupFunc  = false;
  staticLookupData  = false;
  lengthSwitch      = true;
  includeString     = true;
  debug             = false;
}

void HashGenerator::parseEntry(int aLineNo, const char* aBuf) {
  char key[maxLine];
  int keyLen;
  if (!strParse(aBuf, key, keyLen)) {
    die("line %d: Key syntax", aLineNo);
  }
  key[keyLen] = 0;
  if (keyLen == 0) {
    die("line %d: Empty keys are not allowed", aLineNo);
  }
  while (isspace(*aBuf)) {
    aBuf++;
  }
  if (*aBuf == ',') {
    aBuf++;
    while (isspace(*aBuf)) {
      aBuf++;
    }
  }
  add(key, keyLen, aBuf);
}

void HashGenerator::parseOption(int aLineNo, const char* aBuf) {
  char buf[maxLine];
  strcpy(buf, aBuf);
  strTrim(buf);
  char* value = strchr(buf, '=');
  if (value) { // `foo=bar'
    *value++ = 0;
    strTrim(buf);
    strTrim(value);
    setOption(aLineNo, buf, value);
  }
  else if (strncmp(buf, "no-", 3) == 0) { // `no-foo' is the same as `foo=no'
    setOption(aLineNo, buf + 3, "0");
  }
  else { // `foo' is the same as `foo=yes'
    setOption(aLineNo, buf, "1");
  }
}

// Generates its own option lookup source, which turned out
// to be a real pain in the neck when adding/changing options.

#ifdef BOOTSTRAP

#include "optbootstrap.h"

#else

#include "opthash.h"

static inline int optId(const char* aName) {
  OptionEntry* opt = OptionLookup::lookup(aName, strlen(aName));
  return opt ? opt->id : -1;
}

#endif

void HashGenerator::setOption(
  int aLineNo,
  const char* aName,
  const char* aValue)
{
  char buf[128];
  switch (optId(aName)) {
    case optOptimizeHits:
      if (strToBool(aValue, optimizeHits)) return;
      break;
    case optLengthSwitch:
      if (strToBool(aValue, lengthSwitch)) return;
      break;
    case optRandom:
      if (strToBool(aValue, random)) return;
      break;
    case optRandomize:
      if (strToBool(aValue, randomize)) {
        random = randomize;
        return;
      }
      break;
    case optSize:
      if (strToInt(aValue, 0, 9, sizeFactor)) return;
      break;
    case optMinExpectedChar:
      if (strToInt(aValue, 0, 255, minExpectedChar)) return;
      break;
    case optMaxExpectedChar:
      if (strToInt(aValue, 0, 255, maxExpectedChar)) return;
      break;
    case optSource:
      outputSourceName = aValue;
      return;
    case optHeader:
      outputHeaderName = aValue;
      return;
    case optOutput:
      outputSourceName = aValue;
      outputHeaderName = aValue;
      return;
    case optTest:
      outputTestName = aValue;
      return;
    case optLanguage:
      strncpy(buf, aValue, sizeof(buf) - 1);
      buf[sizeof(buf) - 1] = 0;
      strUpper(buf);
      if      (strncmp(buf, "K&R",  3) == 0) language = langOldC;
      else if (strncmp(buf, "KR",   2) == 0) language = langOldC;
      else if (strcmp (buf, "C++"    ) == 0) language = langCPlusPlus;
      else if (strncmp(buf, "ANSI", 4) == 0) language = langAnsiC;
      else if (strcmp (buf, "C"      ) == 0) language = langAnsiC;
      else {
        log("warning: Unknown language, assuming ANSI C\n");
        language = langAnsiC;
      }
      return;
    case optNamespaceName:
      if (*aValue == 0) break;
      namespaceName = aValue;
      return;
    case optEntryStructName:
      if (*aValue == 0) break;
      entryStructName = aValue;
      return;
    case optKeyName:
      if (*aValue == 0) break;
      keyName = aValue;
      return;
    case optLookupStructName:
      if (*aValue == 0) break;
      lookupStructName = aValue;
      return;
    case optLookupDataName:
      if (*aValue == 0) break;
      lookupDataName = aValue;
      return;
    case optLookupFuncName:
      if (*aValue == 0) break;
      lookupFuncName = aValue;
      return;
    case optHashTabName:
      if (*aValue == 0) break;
      hashTabName = aValue;
      return;
    case optKeyTabName:
      if (*aValue == 0) break;
      keyTabName = aValue;
      return;
    case optWeightTabName:
      if (*aValue == 0) break;
      weightTabName = aValue;
      return;
    case optLengthTabName:
      if (*aValue == 0) break;
      lengthTabName = aValue;
      return;
    case optCollisionTabName:
      if (*aValue == 0) break;
      collisionTabName = aValue;
      return;
    case optInlineLookupFunc:
      if (strToBool(aValue, inlineLookupFunc)) return;
      break;
    case optStaticLookupData:
      if (strToBool(aValue, staticLookupData)) return;
      break;
    case optIncludeString:
      if (strToBool(aValue, includeString)) return;
      break;
    case optDebug:
      if (strToBool(aValue, debug)) return;
      break;
    default:
      if (aLineNo) {
        die("line %d: Unknown option `%s'", aLineNo, aName);
      }
      else {
        die("Unknown option `%s'", aName);
      }
  }
  if (aLineNo) {
    die("line %d: Invalid `%s' value", aLineNo, aName);
  }
  else {
    die("Invalid `%s' value", aName);
  }
}

void HashGenerator::openInput() {
  if (inputFileName.empty()) {
    inputStream = &cin;
  }
  else {
    inputFile.open(inputFileName.c_str());
    inputStream = &inputFile;
  }
}

void HashGenerator::closeInput() {
  if (inputStream != &cin) {
    inputFile.close();
    inputStream = &cin;
  }
}

static bool isBlank(const char* aStr) {
  for ( ; ; aStr++) {
    switch (*aStr) {
      case ' ' :
      case '\t': continue;
      case '\0': return true;
      default  : return false;
    }
  }
}

static bool isComment(const char* aStr) {
  return *aStr == '#';
}

static bool getSection(const char* aStr, char* aName) {
  if (*aStr++ != '[') return false;
  for ( ; *aStr && (*aStr != ']'); aStr++) {
    *aName++ = *aStr;
  }
  *aName = 0;
  return true;
}

enum {
  stateEntries,
  stateOptions,
  stateDeclarations,
  stateEntryStruct
};

void HashGenerator::parseInput() {
  parseArgs();
  minKeyLen = INT_MAX;
  maxKeyLen = 0;
  int lineNo = 0;
  char buf[maxLine];
  char section[maxLine];
  int state = stateEntries;
  openInput();
  while (inputStream->getline(buf, sizeof(buf))) {
    lineNo++;
    if (getSection(buf, section)) {
      if      (strcmp(section, "entries"     ) == 0) state = stateEntries;
      else if (strcmp(section, "options"     ) == 0) state = stateOptions;
      else if (strcmp(section, "declarations") == 0) state = stateDeclarations;
      else {
        die("line %d: Invalid section name [%s]", lineNo, section);
      }
    }
    else { // inside a section
      switch (state) {
        case stateEntries:
          if (isBlank(buf) || isComment(buf)) continue;
          parseEntry(lineNo, buf);
          break;
        case stateOptions:
          if (isBlank(buf) || isComment(buf)) continue;
          parseOption(lineNo, buf);
          break;
        case stateDeclarations:
          if (declarations.empty() && isBlank(buf)) continue;
          declarations += buf;
          declarations += '\n';
          break;
      }
    }
  }
  closeInput();
  if (keyCount == 0) {
    die("No entries found");
  }
  minHashKeyLen = min(minKeyLen, maxHashKey+0);
  maxHashKeyLen = min(maxKeyLen, maxHashKey+0);
  for (int i = 0; i < maxHashKey; i++) {
    hashPosMap[i] = i;
  }
  log("%d keys, shortest %d, longest %d\n", keyCount, minKeyLen, maxKeyLen);
  if (minKeyLen == 0) {
    die("Empty keys are not allowed");
  }
}

static Option options[] = {
  {'H', "optimize-hits",       Option::argBoolean},
  {'r', "random",              Option::argBoolean},
  {'R', "randomize",           Option::argBoolean},
  {'N', "size",                Option::argRequired},
  {'S', "length-switch",       Option::argBoolean},
  {'m', "min-expected-char",   Option::argRequired},
  {'M', "max-expected-char",   Option::argRequired},
  {'z', "zero-term",           Option::argBoolean},
  {'o', "output",              Option::argRequired},
  {'c', "source",              Option::argRequired},
  {'h', "header",              Option::argRequired},
  {'T', "test",                Option::argRequired},
  {'e', "entry-struct-name",   Option::argRequired},
  {'k', "key-name",            Option::argRequired},
  {'s', "lookup-struct-name",  Option::argRequired},
  {'d', "lookup-data-name",    Option::argRequired},
  {'f', "lookup-func-name",    Option::argRequired},
  {' ', "key-table-name",      Option::argRequired},
  {' ', "hash-table-name",     Option::argRequired},
  {' ', "weight-table-name",   Option::argRequired},
  {' ', "length-table-name",   Option::argRequired},
  {' ', "weight-table-name",   Option::argRequired},
  {' ', "collision-table-name",Option::argRequired},
  {'I', "inline-lookup-func",  Option::argBoolean},
  {'t', "static-lookup-data",  Option::argBoolean},
  {'i', "include-string-h",    Option::argBoolean},
  {'l', "language",            Option::argRequired},
  {'D', "debug",               Option::argBoolean},
  {' ', "version"},
  {' ', "help"},
  {Option::nonOption},
  {Option::end}
};

void HashGenerator::showUsage() {
cout <<
"The `hashgen' utility generates C/C++ source of a lookup function for a\n"
"given static keyset. Each keyset's entry has a unique key (either string\n"
"or binary) and arbitrary user-defined attributes associated with it.\n"
"The generated lookup function uses hashing, it will find any key in\n"
"constant time using exactly one comparison.\n\n"

"Usage: hashgen [OPTION]... [INPUT-FILE]\n\n"

"The input file contains the keys and attributes, options, and declarations;\n"
"the options may also be specified on the command-line. Only one file can\n"
"be handled at a time. If the input file is not specified, the standard\n"
"input is read. The generated code is written to the standard output by\n"
"default.\n\n"

"Options:\n\n"

" -l, --language=LANG            Language of the generated code: C++,\n"
"                                ANSI C, or K&R C. The default is ANSI C.\n"
" -o, --output=FILE              Output file name, by default to stdout.\n"
"                                May be overriden with -c and -h.\n"
" -h, --header=FILE              Lookup header file name.\n"
" -c, --source=FILE              Lookup source file name. Use with -h to\n"
"                                separate the implementation and the header.\n"
" -T, --test=FILE                Lookup test source file name.\n"
"\n"
" -n, --namespace-name=NAME      C++ namespace name. If the value is empty\n"
"                                (default) the global namespace is used.\n"
" -e, --entry-struct-name=NAME   Key table entry struct name.\n"
" -k, --key-name=NAME            Entry struct key field name.\n"
" -s, --lookup-struct-name=NAME  Lookup struct/class name.\n"
" -d, --lookup-data-name=NAME    Lookup data/instance name.\n"
" -f, --lookup-func-name=NAME    Lookup function name.\n"
" -I, --inline-lookup-func       Make the lookup function inline.\n"
" -t, --static-lookup-data       Hide lookup data within the function\n"
"                                (only if the language is C.)\n"
" -m, --min-expected-char=CODE   Min char code expected in a key.\n"
" -M, --max-expected-char=CODE   Max char code expected in a key.\n"
"                                The default values are 0x01 and 0xFF.\n"
" -z-, --no-zero-term            Treat keys as binary byte arrays\n"
"                                (use memcmp() instead of strcmp()).\n"
" -i-, --no-include-string-h     Don't include <string.h>\n"
"                                or <cstring>.\n"
" -N, --size=SIZE                Hash table size.\n"
"                                The default value is 0.\n"
" -H, --optimize-hits            Faster hits, slower misses.\n"
" -r, --random                   Use rand(3).\n"
" -R, --randomize                Use srand(3).\n"
"                                Assumes --random as well.\n"
" -D, --debug                    Debug output.\n"
"     --version                  Print program's version.\n"
"     --help                     Print this text.\n"
"\n";
}

void HashGenerator::parseArgs() {
  OptionParser parser(options, args);
  const Option* opt;
  const char* arg;
  while (parser.get(opt, arg)) {
    if (opt->id == Option::nonOption) {
      inputFileName = arg;
    }
    else if (strcmp(opt->name, "help") == 0) {
      showUsage();
      exit(0);
    }
    else if (strcmp(opt->name, "version") == 0) {
      // VERSION is expected to be defined in the compiler's command line.
      #ifndef VERSION
      #  define VERSION "(version unknown)"
      #endif
      cout << "hashgen " VERSION "\n";
      exit(0);
    }
    else {
      setOption(0, opt->name, arg);
    }
  }
  if (parser.getError()) {
    die(parser.getErrorMessage());
  }
}

