#include <cstring>
#include <cassert>
#include <algorithm>

#include "utils.h"
#include "hashgen.h"

using namespace std;


// Each bit of the return value is set if the two keys differ
// in the corresponding position (LSB first).

uint HashGenerator::keyDifMask(
  const byte* aKey1, int aLen1,
  const byte* aKey2, int aLen2
) {
  int minLen, maxLen;
  if (aLen1 < aLen2) {
    minLen = aLen1;
    maxLen = aLen2;
  }
  else {
    minLen = aLen2;
    maxLen = aLen1;
  }
  uint dif = 0;
  uint posMask = 1;
  int pos = 0;
  for( ; pos < minLen; pos++, posMask <<= 1) {
    if (aKey1[pos] != aKey2[pos]) {
      dif |= posMask;
    }
  }
  for( ; pos < maxLen; pos++, posMask <<= 1) {
    dif |= posMask;
  }
  return dif;
}

// A bit in the return value is set if the corresponding
// position (LSB first) is redundunt.
// Any two keys must differ in at least one position.
// A position is considered essential if its deletion
// in all keys will result in collision(s).
// Also counts duplicates.

uint HashGenerator::findRedundantPositions() {
  return
    (keyCount < 3000) ?
    findRedundantPositionsFast() :  // ~18Mb
    findRedundantPositionsSlow();   // nothing
}

struct containsBits {
  uint mask, value;
  explicit containsBits(uint aMask, uint aValue) : mask(aMask), value(aValue) {}
  bool operator()(uint aValue) const { return (aValue & mask) == value; }
};

// Uses lots of RAM (keyCount * keyCount * 2 bytes).

uint HashGenerator::findRedundantPositionsFast() {
  // Calculate dif masks for every pair of keys, count duplicates.
  duplicateCount = 0;
  uint* dif = new uint[(keyCount - 1) * keyCount / 2];
  uint* endDif = dif;
  for (Key* key1 = keys; key1 != end; key1++) {
    for (Key* key2 = key1 + 1; key2 != end; key2++) {
      uint difMask = keyDifMask(
        key1->key, key1->keyLen,
        key2->key, key2->keyLen
      );
      *endDif++ = difMask;
      if (difMask == 0) {
        duplicateCount++;
        char buf[maxLine];
        strQuote(reinterpret_cast<const char*>(key1->key), key1->keyLen, buf);
        log("warning: duplicate key %s\n", buf);
      }
    }
  }
  // Find redundant positions.
  uint redundant = 0;
  if (duplicateCount == 0) {
    uint maxPosMask = 1 << (min(maxKeyLen, maxHashKey+0) - 1);
    for (uint posMask = maxPosMask; posMask; posMask >>= 1) {
      if (find_if(dif, endDif, containsBits(~redundant, posMask)) == endDif) {
        redundant |= posMask;
      }
    }
  }
  delete[] dif;
  return redundant;
}

// Mach slower that the fast version but doesn't use RAM at all.

uint HashGenerator::findRedundantPositionsSlow() {
  duplicateCount = 0;
  uint redundant = 0;
  uint maxPosMask = 1 << (min(maxKeyLen, maxHashKey+0) - 1);
  for (uint posMask = maxPosMask ; posMask; posMask >>= 1) {
    bool found = false;
    for (Key* key1 = keys; (key1 != end) && !found; key1++) {
      for (Key* key2 = key1 + 1; key2 != end; key2++) {
        uint difMask = keyDifMask(
          key1->key, key1->keyLen,
          key2->key, key2->keyLen
        );
        if ((difMask == 0) && (posMask == maxPosMask)) {
          // Counts duplicates on the first pass.
          duplicateCount++;
          char buf[maxLine];
          strQuote((const char*)(key1->key), key1->keyLen, buf);
          log("warning: duplicate key %s\n", buf);
        }
        if ((difMask & ~redundant) == posMask) {
          found = true;
          if (posMask != maxPosMask) { // Let it count duplicates.
            break;
          }
        }
      }
    }
    if (duplicateCount) {
      return 0; // Don't bother, the keyset is bogus anyway.
    }
    if (!found) {
      redundant |= posMask;
    }
  }
  return redundant;
}

// Make hash keys. A hash key contains the characters of the key
// that are not redundant. No more than first maxHashKeyLen
// characters may be used. A sorted hash key contains the same
// chars as the hash key, but is sorted.
// Both hashKey and sortedHashKey are redundant (the ``key'' could
// have been used) and needed only for speed.

void HashGenerator::makeHashKeys() {
  uint redundant = findRedundantPositions();
  minHashKeyLen = INT_MAX;
  maxHashKeyLen = 0;
  for (Key* key = keys; key != end; key++) {
    int maxLen = min(key->keyLen, maxHashKey+0);
    int hashKeyLen = 0;
    short buf[maxHashKey];
    uint posMask = 1;
    for (int keyPos = 0; keyPos < maxLen; keyPos++, posMask <<= 1) {
      if (redundant & posMask) continue;
      buf[hashKeyLen++] = key->key[keyPos];
    }
    key->hashKey = new short[hashKeyLen];
    key->sortedHashKey = new short[hashKeyLen];
    copy(buf, buf + hashKeyLen, key->hashKey);
    sort(buf, buf + hashKeyLen);
    copy(buf, buf + hashKeyLen, key->sortedHashKey);
    key->hashKeyLen = hashKeyLen;
    if (hashKeyLen > maxHashKeyLen) {
      maxHashKeyLen = hashKeyLen;
    }
    if (hashKeyLen < minHashKeyLen) {
      minHashKeyLen = hashKeyLen;
    }
  }
  // Adjust mapping from hash key pos to key pos.
  int hashKeyPos = 0;
  uint posMask = 1;
  for (int keyPos = 0; keyPos < maxHashKey; keyPos++, posMask <<= 1) {
    if (redundant & posMask) continue;
    hashPosMap[hashKeyPos++] = hashPosMap[keyPos];
  }
  // Report results.
  log("shortest hash key %d, longest %d\n", minHashKeyLen, maxHashKeyLen);
  log("hash positions ");
  for (int pos = 0; pos < maxHashKeyLen; pos++) {
    log("%d%c", hashPosMap[pos], (pos == maxHashKeyLen - 1) ? '\n' : ',');
  }
}

// Function object for find_if'ing a key "similar" to the given key.
// ("similar" means that its hash key contains the same chars (may be
// in different order) and the key lengths are the same (length is used
// for hashing).

struct similarHashKey {
  const short* sortedHashKey;
  int keyLen;
  int hashKeyLen;
  explicit similarHashKey(const Key* aKey)
  : sortedHashKey(aKey->sortedHashKey),
    keyLen(aKey->keyLen),
    hashKeyLen(aKey->hashKeyLen)
  {}
  bool operator()(const Key& aKey) const {
    return
      (keyLen == aKey.keyLen) &&
      (hashKeyLen == aKey.hashKeyLen) &&
      equal(sortedHashKey, sortedHashKey + hashKeyLen, aKey.sortedHashKey);
  }
};

// True if there are at least two keys in the keyset with the same
// set of characters in their hash keys *and* the same key lengths.

bool HashGenerator::findSimilarHashKeys() {
  for (Key* key = keys; key != end; key++) {
    if (find_if(keys, key, similarHashKey(key)) != key) {
      return true;
    }
  }
  return false;
}

// Offsets characters of all hash keys by offsets[].
// Re-sorts sorted hash keys.

void HashGenerator::applyOffsets() {
  for (Key* key = keys; key != end; key++) {
    int len = key->hashKeyLen;
    for (int pos = 0; pos < len; pos++) {
      key->hashKey[pos] = key->key[hashPosMap[pos]] + offsets[pos];
    }
    copy(key->hashKey, key->hashKey + len, key->sortedHashKey);
    sort(key->sortedHashKey, key->sortedHashKey + len);
  }
}

// Resolves the "if/fi" problem by offseting chracters so that they
// depend on their position (a char in some position is different
// from any char in any other position).

void HashGenerator::computeCharOffsets() {
  minOffset = 0;
  maxOffset = 0;
  memset(offsets, 0, sizeof(offsets));
  if (duplicateCount) return; // Don't bother, collisions will be there anyway.
  if (!findSimilarHashKeys()) return; // Don't need offseting.
  // Compute char code ranges for every key position.
  struct {
    short minChar;
    short maxChar;
  } ranges[maxHashKey];
  for (int pos = 0; pos < maxHashKeyLen; pos++) {
    ranges[pos].minChar = maxCharCount;
    ranges[pos].maxChar = 0;
  }
  for (Key* key = keys; key != end; key++) {
    for (int pos = 0; pos < key->hashKeyLen; pos++) {
      short c = key->hashKey[pos];
      ranges[pos].maxChar = max(ranges[pos].maxChar, c);
      ranges[pos].minChar = min(ranges[pos].minChar, c);
    }
  }
  // Compute offsets for every hash key position so that char code
  // ranges at different positions don't overlap.
  offsets[0] = 0;
  for (int pos = 1; pos < maxHashKeyLen; pos++) {
    int offset = offsets[pos - 1];
    offset += ranges[pos - 1].maxChar - ranges[pos].minChar + 1;
    offset = max(offset, -minExpectedChar); // char code must be >= 0
    minOffset = min(minOffset, offset);
    maxOffset = max(maxOffset, offset);
    offsets[pos] = offset;
  }
  // Offset hash keys.
  applyOffsets();
  if (debug) {
    log("offsets ");
    for (int pos = 0; pos < maxHashKeyLen; pos++) {
      log("%d%c", offsets[pos], (pos == maxHashKeyLen - 1) ? '\n' : ',');
    }
  }
}

void HashGenerator::checkExpectedCharRange() {
  int cmin = INT_MAX;
  int cmax = 0;
  for (Key* key = keys; key != end; key++) {
    for (int pos = 0; pos < key->keyLen; pos++) {
      int c = key->key[pos];
      cmin = min(cmin, c);
      cmax = max(cmax, c);
    }
  }
  minExpectedChar = min(minExpectedChar, cmin);
  maxExpectedChar = max(maxExpectedChar, cmax);
  if (minExpectedChar == 0) {
    zeroTerm = false;
  }
}

// For all hash keys (already offset with offsets[]), computes:
// - min and max char code,
// - sums of char occurence counts,
// - char sets.

void HashGenerator::gatherStatistics() {
  minChar = INT_MAX;
  maxChar = 0;
  memset(charFreq, 0, sizeof(charFreq));
  for (Key* key = keys; key != end; key++) {
    for (int pos = 0; pos < key->hashKeyLen; pos++) {
      int c = key->hashKey[pos];
      charFreq[c]++;
      minChar = min(minChar, c);
      maxChar = max(maxChar, c);
    }
  }
  int charSetSize = (maxChar + 8) / 8;
  for (Key* key = keys; key != end; key++) {
    key->freq = 0;
    key->charSet = new byte[charSetSize];
    memset(key->charSet, 0, charSetSize);
    for (int pos = 0; pos < key->hashKeyLen; pos++) {
      int c = key->hashKey[pos];
      key->freq += charFreq[c];
      key->charSet[c >> 3] |= 1 << (c & 7);
    }
  }
  if (debug) {
    log("hash char code min %d, max %d, ", minChar, maxChar);
    log("range %d\n", maxChar + 1 - minChar);
  }
}

