#include <cstdlib>
#include <ctime>
#include "hashgen.h"
#include "utils.h"

using namespace std;


#define commentTest \
  "Test for hashed lookup code generated by hashgen " VERSION ".\n" \
  "Measures the average time of both successful and unsuccessful lookups." 

#define commentTestHits \
  "Hit test. If/goto overhead is negligible, so it's okay to test for\n" \
  "both correctness and speed at the same time.  There is no sense in\n" \
  "checking the key pointed to  by the lookup function's return value\n" \
  "after a successful search, since  the only place where  a non-null\n" \
  "pointer is returned is right after successful strcmp() or memcmp()."

#define commentTestMisses \
  "Miss test.  Characters of the test keys are random within range\n" \
  "from min-expected-char to max-expected-char, their length from 0\n" \
  "to a bit more than maximum length of the existing keys.\n" \
  "The results should not be taken too seriously, the keys are too\n" \
  "random, so strcmp() or memcmp() is not called most of the time\n" \
  "because the calculated hash values are out of range."

void HashGenerator::generateTestBanner() {
  generateComment(commentTest);
  out("\n\n");
}

void HashGenerator::generateTestIncludes() {
  if (language == langCPlusPlus) {
    out("#include <ctime>\n");
    out("#include <iostream>\n");
    out("#include <iomanip>\n\n");
    out("using namespace std;\n");
    if (!namespaceName.empty()) {
      out("using namespace %s;\n", namespaceName.c_str());
    }
  }
  else {
    out("#include <time.h>\n");
    out("#include <stdio.h>\n");
  }
  out("\n");
  out("#ifndef CLK_TCK\n");
  out("#define CLK_TCK CLOCKS_PER_SEC\n");
  out("#endif\n\n");
}

void HashGenerator::generateTestErrorBranch(
  const char* aLabel,
  const char* aMessage)
{
  generateLabel(aLabel);
  if (language == langCPlusPlus) {
    out("cerr << \"%s: %s\\n\";\n", outputSourceName.c_str(), aMessage);
  }
  else {
    out("printf(\"%s: %s\\n\");\n", outputSourceName.c_str(), aMessage);
  }
  out("exit(1);\n");
}

void HashGenerator::generateTestKey(
  const byte* aKey,
  int aLen,
  bool aExisting)
{
  out("if (");
  if (language == langCPlusPlus) {
    out("%s::", lookupStructName.c_str());
  }
  char quoted[maxLine];
  strQuote(reinterpret_cast<const char*>(aKey), aLen, quoted);
  out("%s(%s, %d)", lookupFuncName.c_str(), quoted, aLen);
  if (aExisting) {
    out((language == langCPlusPlus) ? " == 0" : " == NULL");
  }
  out(") goto %s;\n", aExisting ? "err_miss" : "err_hit");
}

void HashGenerator::generateTestKeys(bool aExisting) {
  const double minTestTime = 0.5;
  int testIterations = 1000000 / keyCount;
  srand(static_cast<unsigned>(time(0)));
  generateComment(aExisting ? commentTestHits : commentTestMisses);
  out("\n");
  out("count = 0;\n");
  out("tick = clock();\n");
  out("do ");
  startScope();
  out("for (i = 0; i < %d; i++) ", testIterations);
  startScope();
  for (Key* key = keys; key != end; key++) {
    byte buf[maxLine];
    byte* testKey = key->key;
    int testLen = key->keyLen;
    if (!aExisting) {
      testKey = buf;
      testLen = rand() % (maxKeyLen + 2);
      int charRange = maxExpectedChar + 1 - minExpectedChar;
      bool found = false;
      do {
        byte* p = buf;
        for (int i = 0; i < testLen; i++) {
          *p++ = minExpectedChar + rand() % charRange;
        }
        *p = 0;
        found = false;
        for (Key* k = keys; k != end; k++) {
          if (k->keyLen != testLen) continue;
          if (memcmp(k->key, testKey, testLen) != 0) continue;
          found = true;
          break;
        }
      } while (found);
    }
    generateTestKey(testKey, testLen, aExisting);
  }
  endScope(1);
  out("count++;\n");
  out("time = (clock() - tick) / (double)CLK_TCK;\n");
  endScope();
  out(" while ((time < %0.1f) && (count < 1000000));\n", minTestTime);
  if (language == langCPlusPlus) {
    out("cout\n");
    out("<< \"");
    out(
      aExisting ?
      "successful search time " :
      "search failure time    "
    );
    out("\"\n");
    out("<< setw(6)\n");
    out("<< int(time * 1e9 / %d / %d / count)\n", testIterations, keyCount);
    out("<< \"ns (\" << setw(0)\n");
    out("<< int(%d * count / time)\n", testIterations * keyCount);
    out("<< \" searches/sec)\\n\";\n");
  }
  else {
    out("printf(\"");
    out(aExisting ? "successful search time " : "search failure time    ");
    out("\");\n");
    out("printf(\"%%6.0fns \", ");
    out("time * 1e9 / %d / %d / count);\n", testIterations, keyCount );
    out("printf(\"(%%d searches/sec)\\n\", ");
    out("(int)(%d * count / time));\n", testIterations * keyCount);
  }
}

void HashGenerator::generateTestMain() {
  out("int main() ");
  startScope();
  out("clock_t tick;\n");
  out("double time;\n");
  out("unsigned count;\n");
  out("unsigned i;\n");
  generateTestKeys(true);
  generateTestKeys(false);
  out("return 0;\n");
  generateTestErrorBranch("err_hit",  "non-existing key found");
  generateTestErrorBranch("err_miss", "key not found");
  endScope(1);
}

void HashGenerator::generateTestCode() {
  if (outputHeaderName.empty()) {
    string saveOutputHeaderName = outputHeaderName;
    string saveOutputSourceName = outputSourceName;
    outputSourceName = outputTestName;
    outputHeaderName = outputTestName;
    openOutput(outputTestName);
    generateTestBanner();
    doGenerateLookupCode();
    string outputHeaderName = saveOutputHeaderName;
    string outputSourceName = saveOutputSourceName;
  }
  else {
    openOutput(outputTestName);
    generateTestBanner();
    out("#include \"%s\"\n", outputHeaderName.c_str());
  }
  generateTestIncludes();
  generateTestMain();
  closeOutput();
}

